"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.updateDecorations = exports.onUpdatedAST = exports.activate = void 0;
const vscode = require("vscode");
const typescript_estree_1 = require("@typescript-eslint/typescript-estree");
const enterExitTraverser_1 = require("./enterExitTraverser");
function activate(context) {
    let status = vscode.window.createStatusBarItem(vscode.StatusBarAlignment.Left, 0);
    let recentRunLength = 10 * 1000;
    let recentRuns = [];
    async function timeCode(code) {
        let runStart = Date.now();
        try {
            await code();
        }
        catch (e) {
            (vscode.window.showInformationMessage(`Parse error ${e.stack}!`));
        }
        finally {
            let runEnd = Date.now();
            recentRuns.push({ start: runStart, end: runEnd });
        }
        updateUsageFraction();
    }
    function updateUsageFraction() {
        let sum = 0;
        let threshold = Date.now() - recentRunLength;
        for (let i = recentRuns.length - 1; i >= 0; i--) {
            let run = recentRuns[i];
            if (run.start < threshold) {
                recentRuns.splice(i);
            }
            let time = run.end - run.start;
            sum += time;
        }
        let usageFrac = sum / recentRunLength;
        let lastTime = 0;
        if (recentRuns.length > 0) {
            lastTime = recentRuns[recentRuns.length - 1].end - recentRuns[recentRuns.length - 1].start;
        }
        status.text = `Closer ${sum}ms/${recentRunLength}ms, ${recentRuns.length}, Last ${lastTime}ms`;
        status.show();
    }
    let timeout = undefined;
    let updateStatus = "none";
    triggerUpdateDecorations();
    function triggerUpdateDecorations() {
        // Restart while the user is typing
        if (timeout) {
            clearTimeout(timeout);
            timeout = undefined;
        }
        else {
            if (updateStatus === "pending") {
                updateStatus = "pendingStale";
                return;
            }
            if (updateStatus === "pendingStale") {
                return;
            }
            updateStatus = "pending";
        }
        timeout = setTimeout(async () => {
            timeout = undefined;
            try {
                const activeEditor = vscode.window.activeTextEditor;
                if (!activeEditor)
                    return;
                let languageId = activeEditor.document.languageId;
                if (languageId !== "typescriptreact" && languageId !== "typescript" && languageId !== "javascript" && languageId !== "javascriptreact")
                    return;
                await timeCode(() => updateDecorations(activeEditor));
            }
            finally {
                let status = updateStatus;
                updateStatus = "none";
                if (status === "pendingStale") {
                    triggerUpdateDecorations();
                }
            }
        }, 500);
    }
    vscode.window.onDidChangeActiveTextEditor(editor => {
        triggerUpdateDecorations();
    }, null, context.subscriptions);
    vscode.workspace.onDidChangeTextDocument(event => {
        var _a;
        if (event.document === ((_a = vscode.window.activeTextEditor) === null || _a === void 0 ? void 0 : _a.document)) {
            triggerUpdateDecorations();
        }
    }, null, context.subscriptions);
    context.subscriptions.push(status);
}
exports.activate = activate;
let onUpdatedCallbacks = [];
function onUpdatedAST(updatedCallback) {
    onUpdatedCallbacks.push(updatedCallback);
}
exports.onUpdatedAST = onUpdatedAST;
function splitRenderOptions(intersection) {
    if ("node" in intersection) {
        let { node, ...remaining } = intersection;
        let { start, end } = node.loc;
        intersection = Object.assign(remaining, { range: new vscode.Range(start.line - 1, start.column, end.line - 1, end.column) });
    }
    let { light, dark, before, after, ...remaining } = intersection;
    return {
        options: remaining,
        instance: {
            light,
            dark,
            before,
            after
        },
        range: intersection.range,
        hoverMessage: intersection.hoverMessage,
    };
}
let currentDecorations = new Map();
async function updateDecorations(editor) {
    let doc = editor.document;
    let ast = typescript_estree_1.parse(doc.getText(), {
        module: true,
        ts: true,
        jsx: true,
        next: true,
        loc: true,
        ranges: true,
        raw: true,
    });
    let decorations = [];
    for (let updatedCallback of onUpdatedCallbacks) {
        let callbacksExpired = false;
        function setDecoration(decoration) {
            // TODO: Add support for this, so they can asynchronously trigger updates, via info from a remote source
            //  (although... the callback is async, so usually they can get the data before the callback ends. This is mostly
            //  so you can watch a remote value and continuously update the the UI. WHICH, also really requires the ability to remove
            //  decorations, or even better to explicitly update them (which should be more efficient anyway)).
            if (callbacksExpired)
                throw new Error(`Cannot call callback after update callback finishes.`);
            decorations.push(decoration);
        }
        function traverse(config) {
            if (callbacksExpired)
                throw new Error(`Cannot call callback after update callback finishes.`);
            new enterExitTraverser_1.EnterExitTraverser({
                enter: config.enter,
                exit: config.exit,
            }).traverse(ast);
        }
        await updatedCallback({
            doc,
            ast,
            setDecoration,
            traverse
        });
        callbacksExpired = true;
    }
    let decMap = new Map();
    for (let dec of decorations) {
        let { options, instance, range, hoverMessage } = splitRenderOptions(dec);
        let optionsJSON = JSON.stringify(options);
        let decObj = decMap.get(optionsJSON);
        if (!decObj) {
            decObj = {
                options,
                instances: []
            };
            decMap.set(optionsJSON, decObj);
        }
        decObj.instances.push({
            renderOptions: instance,
            range,
            hoverMessage: hoverMessage,
        });
    }
    for (let [optionsJSON, dec] of currentDecorations) {
        if (!decMap.has(optionsJSON)) {
            editor.setDecorations(dec, []);
            currentDecorations.delete(optionsJSON);
        }
    }
    for (let [optionsJSON, decObj] of decMap) {
        let dec = currentDecorations.get(optionsJSON);
        if (!dec) {
            dec = vscode.window.createTextEditorDecorationType(decObj.options);
            currentDecorations.set(optionsJSON, dec);
        }
        editor.setDecorations(dec, decObj.instances);
    }
}
exports.updateDecorations = updateDecorations;
require("./decorationSetup");
//# sourceMappingURL=extension.js.map